-- note: investigate use of instance Read instead

module RegExpData (
    RegExp(..), allExp, ewp,
    isConcat, isUnion, isKleene,
    getUnion, getConcat, getKleene) where

import Prelude
import General

data RegExp = Omega
            | Terminal Char
            | Kleene RegExp
            | Concat [RegExp]
            | Union [RegExp]
            deriving (Show)


isConcat (Concat _) = True
isConcat _          = False

isUnion (Union _) = True
isUnion _         = False

isKleene (Kleene _) = True
isKleene _          = False

getConcat (Concat x) = x
getUnion  (Union  x) = x
getKleene (Kleene x) = x

lambda = Kleene Omega


{-
display a regular expression as a piece of text, done this way to make it useful
in both the program and during development

Aim is to show it with the minimal amount of brackets
Kleene Omega is automatically transformed into "^", representing Lamdba
-}

{-
implement an ordered Precedence relationship
a < b means that if a regular expression of type a is combined under b it must be bracketed
-}
data Order =  OUnion | OConcat | OKleene
    deriving (Eq, Ord)

{-
instance Show RegExp where
    show x = f OUnion x
        where
            f :: Order -> RegExp -> String
            
            f p (Terminal x  ) = [x]
            f p (Omega       ) = "0"
            f p (Kleene Omega) = "^"
            
            f p (Kleene x) = f OKleene x ++ "*"
            f p (Concat xs) = brack (OConcat < p) (foldr join "" xs)
                where join x xs  = f OConcat x ++ xs
            
            f p (Union x) = brack (OUnion < p) (tail (foldr join "" x))
                where join x xs = '+' : show x ++ xs
                
            brack False s = s
            brack True  s = '(' : s ++ ")"
-}

instance Eq RegExp where
    Omega      == Omega      = True
    Terminal x == Terminal y = x == y
    Kleene   x == Kleene   y = x == y
    Concat   x == Concat   y = x == y
    Union    x == Union    y = setEq x y
    _          == _          = False

    
allExp :: (RegExp -> RegExp) -> RegExp -> RegExp
allExp f (Concat xs) = f (Concat (map (allExp f) xs) )
allExp f (Union  xs) = f (Union  (map (allExp f) xs) )
allExp f (Kleene xs) = f (Kleene      (allExp f  xs) )
allExp f x           = f x

-- empty word property (is ^ a member of the langauge)
ewp :: RegExp -> Bool
ewp (Concat xs) = and (map ewp xs)
ewp (Union  xs) = or  (map ewp xs)
ewp (Kleene xs) = True
ewp _           = False


{-#--
-- emptiness property (is the langauge equal to 0)
empty :: RegExp -> Bool
empty (Omega   ) = True
empty (Union  a) = and (map empty a)
empty (Concat a) = or  (map empty a)
empty _          = True

-- blank property (is the language equal to ^)
blank :: RegExp -> Bool
blank (Omega       ) = False
blank (Terminal   _) = False
blank (Kleene Omega) = True
blank (Kleene     a) = blank a
blank (Concat     a) = and (map blank a)

-- one must have the blank property, the others may have blank or empty
blank (Union  a) = find_blank False a
    where
        find_blank b []     = b
        find_blank b (x:xs) = (blank_x || empty x) && (find_blank (b || blank_x) xs)
            where blank_x = blank x
--}
